"""IRC parser and utility library."""

from irclib._version import (
    __version__,
    __version_tuple__,
    version,
    version_tuple,
)

__all__ = ["version", "version_tuple", "__version__", "__version_tuple__"]
